#include "Parameters.h"
#include "MemoryAllocators.h"
#include "MathStats.h"
#include "Pedigree.h"
#include "Error.h"
#include <ctime>

void readmarkers(StringHash & markerhash, const char * datfile)
{
	FILE * f = fopen(datfile, "rt");
	
	if (f == NULL)  return;
	
	String buffer;
	StringArray tokens;
	
	printf("Loading list of markers into a stringhash ... \n");
	
	while(!feof(f))
	{
		buffer.ReadLine(f);
		tokens.ReplaceTokens(buffer);
		
		if (tokens.Length() == 0) continue;
		
		if (tokens.Length() != 2)
		   error("Each line should list exactly one flag and one marker name, but\n"
		         "the following line appears to have problem:\n\n"
		         "%s\n", (const char *) buffer);
		
		markerhash.Add(tokens[1]);
	}

	fclose(f);
}


int main(int argc, char ** argv)
{
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

clock_t startt, endt;    
startt = clock();

	String trueped, truedat, impped, impdat, maskped, maskdat, outfile("CalcMatch.out");
	bool match = false, bySNP = true, byPerson = false, byGeno = false, accuracyByGeno = true, maskflag = false;
	printf("CalcMatch 1.1.2 -- Evaluate Imputed Genotypes against True/Experimental Ones\n"
	       "             (c) 2007 Yun Li\n\n");
	
	ParameterList pl;
	
BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("trueped", &trueped)
      LONG_STRINGPARAMETER("truedat", &truedat)
      LONG_STRINGPARAMETER("impped", &impped)
      LONG_STRINGPARAMETER("impdat", &impdat)
      LONG_STRINGPARAMETER("maskped", &maskped)
      LONG_STRINGPARAMETER("maskdat", &maskdat)
   LONG_PARAMETER_GROUP("Output")
      LONG_STRINGPARAMETER("prefix", &outfile)
      LONG_PARAMETER("match", &match)
      LONG_PARAMETER("bySNP", &bySNP)
      LONG_PARAMETER("byPerson", &byPerson)
      LONG_PARAMETER("byGeno", &byGeno)
      LONG_PARAMETER("accuracyByGeno", &accuracyByGeno)
      LONG_PARAMETER("maskflag", &maskflag)
END_LONG_PARAMETERS();
	
	pl.Add(new LongParameters("Available Options", longParameters));
	pl.Add(new HiddenString('o', "Output File", outfile));
	
	pl.Read(argc, argv);
	pl.Status();
	
	// Setup and load pedigrees ...
	Pedigree truth;
	truth.Prepare(truedat);
	truth.Load(trueped);
	StringHash truemarkers;
	StringHash flippedmarkers;
	int numtruthmarker = truth.markerCount;
	printf("Loaded TRUE/EXPERIMENTAL pedigree with:\n"
	       "    %d individuals at %d markers\n\n",
	       truth.count, truth.markerCount);
	
	Pedigree imputed;
	StringHash impmarkers;
	readmarkers(impmarkers, impdat);
	imputed.Prepare(impdat);
	imputed.Load(impped);
	printf("Loaded IMPUTED pedigree with:\n"
	       "    %d individuals at %d markers\n\n",
	       imputed.count, imputed.markerCount);
	
	Pedigree masked;
	if (maskflag)
	{
		masked.Prepare(maskdat);
		masked.Load(maskped);
		printf("Loaded MASKED pedigree with:\n"
	           "    %d individuals at %d markers\n\n",
	           masked.count, masked.markerCount);
	}
	
	// counts at genotypes where masked, but with truth and imputed both available;
	// here allocate memory for markers in truth.dat (stored first b/c first read in truth);
	int * maskedgenos = new int [numtruthmarker];
	int * matchedgenos = new int [numtruthmarker];
	int * matchedalleles = new int [numtruthmarker];
	if (maskedgenos == NULL || matchedgenos == NULL || matchedalleles == NULL) 
	  error("Failed to allocate enough memory for marker-specific errors\n\n");

	int ** maskedbygeno;
	int ** mmbygeno; //mismatch by geno: AAtrue->ABimp; AA-> BB; AB->AA; AB->BB etc;
	if (byGeno || accuracyByGeno)
	{
		maskedbygeno = AllocateIntMatrix(numtruthmarker, 3);
		mmbygeno = AllocateIntMatrix(numtruthmarker, 6);
		if (maskedbygeno == NULL || mmbygeno ==NULL)
			error("Failed to allocate memory for marker-specific bygeno errors\n\n");
	}
	  
	int personmaskedgenos = 0;
	int personmatchedgenos = 0;
	int personmatchedalleles = 0;
	
	for (int j = 0; j < numtruthmarker; j++)
	{
		maskedgenos[j] = 0;
		matchedgenos[j] = 0;
		matchedalleles[j] = 0;
	}

	if (byGeno || accuracyByGeno)
	{
		for (int j = 0; j < numtruthmarker; j++)
		{
			for (int i = 0; i < 3; i++)
				maskedbygeno[j][i] = 0;
			for (int i = 0; i < 6; i++)
				mmbygeno[j][i] = 0;
		}
	}
	
	bool * markerinimp = new bool [numtruthmarker];
	if (markerinimp == NULL)
	  error("Failed to allocate momory for flagging imputed markers\n\n");
	
	for (int j = 0; j < numtruthmarker; j++)
	{
		// check if the marker in imp.dat
		markerinimp[j] = false;
		if (impmarkers.Find(truth.markerNames[j]) >= 0)
		{
			markerinimp[j] = true;
		}
		// printf("%s %d %d\n", (const char *) truth.markerNames[j], impmarkers.Find(truth.markerNames[j]), markerinimp[j]);
		
	}
	
	FILE * mf;
	
	if (match)
	{
  	mf = fopen(outfile + ".match", "wt");
	  if (mf == NULL)
	  {
  		printf("Error opening output file [%s.match]\n", (const char *) outfile);
  		exit(1);
		}
  }
	//printf("famid pid true_geno imp_geno\n");
	
	FILE * personf ;
	if (byPerson)
	{
  	personf = fopen(outfile + ".byPerson", "wt");
	  if (personf == NULL)
	  {
  		printf("Error opening output file [%s.byPerson]\n", (const char *) outfile);
  		exit(1);
		}
		fprintf(personf, "famid subjID gErr aErr matchedG matchedA maskedG\n");
	}
	
	for (int i = 0; i < truth.count; i++)
	{
  	Person * impperson;
  	impperson = imputed.FindPerson(truth[i].famid, truth[i].pid);
  	if (impperson == NULL) continue;
  	
  	if (match)
  	  fprintf(mf, "%s %s ", (const char * ) truth[i].famid, (const char * ) truth[i].pid);
  	  
  	if (byPerson)
  	{
	    personmaskedgenos = 0;
	    personmatchedgenos = 0;
	    personmatchedalleles = 0;
	  }
  	  
		for (int j = 0; j < numtruthmarker; j++)
		{
			Person * maskperson = masked.FindPerson(truth[i].famid, truth[i].pid);
			
			if (! markerinimp[j]) continue;
			
			int maskmarkerid;
			if (maskperson != NULL)
			  maskmarkerid = masked.LookupMarker(truth.markerNames[j]);

			MarkerInfo * info = Pedigree::GetMarkerInfo(truth.markerNames[j]);
			
			// when masked not provided (assume masked) or;
			// the person not found in masked or;
			// masked (i.e., sites = 0/0 etc);
			if ( (! maskflag) || maskperson == NULL || 
			     ( ! (maskperson->markers[maskmarkerid].isKnown())) )
			{
				// truth and imputed both available/known, here truth;
				if ( truth[i].markers[j].isKnown() )
				{
					int impmarkerid = imputed.LookupMarker(truth.markerNames[j]);
					
					// here imputed;
					if ( impperson->markers[impmarkerid].isKnown() )
					{
						maskedgenos[j] ++;
						if (byPerson)
						  personmaskedgenos ++;
						
						// match of alleles, not based on SequenceCoded();
						int t1 = (int) truth[i].markers[j].one;
						int t2 = (int) truth[i].markers[j].two;
						int i1 = (int) impperson->markers[impmarkerid].one;
						int i2 = (int) impperson->markers[impmarkerid].two;
//printf("TRUE: %d/%d \n", t1,t2);
//printf("IMP: %d/%d \n", i1,i2);
//printf("TRUE: %c/%c \n", truth[i].markers[j].one,truth[i].markers[j].two);
//printf("IMP: %c/%c \n", impperson->markers[impmarkerid].one,impperson->markers[impmarkerid].two);
//printf("TRUE: %s/%s \n", (const char *) info->GetAlleleLabel(t1),(const char *) info->GetAlleleLabel(t2));
//printf("IMP: %s/%s \n", (const char *) info->GetAlleleLabel(i1),(const char *) info->GetAlleleLabel(i2));
						
						if (byGeno || accuracyByGeno)
						{
							int tg = truth[i].markers[j].SequenceCoded() - 1;
							int ig = impperson->markers[impmarkerid].SequenceCoded() - 1;
//printf("%d %d\n", tg, ig);
							if (tg < 0 || tg > 2 || ig < 0 || ig > 2) // only for bi-allelic markers;
							{
								if (flippedmarkers.Find(truth.markerNames[j]) == -1)
									flippedmarkers.Add(truth.markerNames[j]);
									
							  printf("WARNING: skipping marker %s with more than two alleles\n", (const char *) truth.markerNames[j]);
							  continue;
							  
						  }

							maskedbygeno[j][tg] ++;

							if (tg == 0 && ig != tg)
								mmbygeno[j][ig - 1] ++;
							if (tg == 1 && ig == 0)
								mmbygeno[j][2] ++;
							if (tg == 1 && ig == 2)
								mmbygeno[j][3] ++;
							if (tg == 2 && ig != tg)
								mmbygeno[j][4+ig] ++;

						}
						// complete match
						if ( (t1 == i1 && t2 == i2) || (t1 == i2 && t2 == i1) )
						{
							matchedgenos[j] ++;
							matchedalleles[j] += 2;
							if (byPerson)
							{
  							personmatchedgenos ++;
  							personmatchedalleles += 2;
							}
							if (match)
							  fprintf(mf, "2 ");
						}
						
						// match only one allele
						else if ( (t1 == i1 && t2 != i2) ||
						          (t1 == i2 && t2 != i1) ||
						          (t2 == i1 && t1 != i2) ||
						          (t2 == i2 && t1 != i1) )
						{
							matchedalleles[j] ++;
							if (byPerson)
							  personmatchedalleles ++;
							if (match)
							  fprintf(mf, "1 ");
						}
						
						// both alleles fail to match
						else
						{
  						if (match)
  						  fprintf(mf, "0 ");
						}
					}
					else //imputed genotype not known
					{
						//printf("%s %s %d/%d NA\n", (const char * ) truth[i].famid, (const char * ) truth[i].pid, truth[i].markers[j].one, truth[i].markers[j].two);
						//printf("%s %s %d/%d NA\n", (const char * ) truth[i].famid, (const char * ) truth[i].pid, truth[i].markers[j].SequenceCoded());
						if (match)
						  fprintf(mf, "NA ");
					}
				}
				else // true genotype not known
				{
  				if (match)
  				  fprintf(mf, "NA ");
				}
			}
			else // when available in masked, i.e., genotype not to be compared 
			{
  			if (match)
  			  fprintf(mf, "NA ");
			}
		}
		if (match)
		  fprintf(mf, "\n");
		  
		if (byPerson)
		{
  		if (personmaskedgenos > 0)
  		{
    		double gErr = 1.00 - (double) personmatchedgenos / (double) personmaskedgenos;
    		double aErr = 1.00 - (double) personmatchedalleles * 0.5 / (double) personmaskedgenos;
    		fprintf(personf, "%s %s %5.4f %5.4f %d %d %d\n", (const char * ) truth[i].famid, (const char * ) truth[i].pid,
    		                 gErr, aErr, personmatchedgenos, personmatchedalleles, personmaskedgenos);
  		}
		}
	}
	
	if (match)
	{
  	mf = fopen(outfile + ".formatch.snps", "wt");
  	if (mf == NULL)
  	{
  		printf("Error opening output file [%s.formatch.snps]\n", (const char *) outfile);
  		exit(1);
  	}
  	for (int j = 0; j < numtruthmarker; j++)
  	{
    	if (! markerinimp[j]) continue;
    	fprintf(mf, "%s\n", (const char *) truth.markerNames[j]);
  	}
	}
	
	FILE * strandf;
	if (byGeno || accuracyByGeno)
	{
		strandf  = fopen(outfile + ".minusstrand", "wt");
		if (strandf == NULL)
	        {
        	        printf("Error opening output file [%s.minusstrand]\n", (const char *) outfile);
                	exit(1);                 
	        }
		fprintf(strandf, "SNP\n");
	}


	FILE * of = fopen(outfile + ".bySNP", "wt");
	if (of == NULL)
	{
		printf("Error opening output file [%s.bySNP]\n", (const char *) outfile);
		exit(1);
	}
	
	fprintf(of, "SNP gErr aErr matchedG matchedA maskedG");
	if (byGeno)
		fprintf(of, " hetAerr AL1 AL2 freq1 MAF #true1/1 mm1/2 mm2/2 #true1/2 mm1/1 mm2/2 #true2/2 mm1/1 mm1/2");
	if (accuracyByGeno)
		fprintf(of, " a1major a2minor freq1 accuracy11 accuracy12 accuracy22");
	fprintf(of, "\n");
	long int ttlmaskedgenos = 0;
	long int ttlmatchedgenos = 0;
	int ttlmatchedalleles = 0;
	
	//for (int j = 0; j < truth.markerCount; j++)
	for (int j = 0; j < numtruthmarker; j++)
	{
		MarkerInfo * info = Pedigree::GetMarkerInfo(truth.markerNames[j]);

		if (! markerinimp[j]) continue;
		
		if (maskedgenos[j] <= 0)
		{
			fprintf(of, "%s NA NA NA NA 0\n", (const char *) truth.markerNames[j]);
		}
		else
		{
			ttlmaskedgenos += maskedgenos[j];
			ttlmatchedgenos += matchedgenos[j];
			ttlmatchedalleles += matchedalleles[j];
			
			double gErr =  1.00 - (double) matchedgenos[j] / (double) maskedgenos[j] ;
			double aErr =  1.00 - (double) matchedalleles[j] * 0.5 / (double) maskedgenos[j];
			fprintf(of, "%s %5.4f %5.4f %d %d %d", (const char *) truth.markerNames[j], 
			                 gErr, aErr, matchedgenos[j], matchedalleles[j], maskedgenos[j]);
			if (byGeno)
			{
				int masked = maskedbygeno[j][0] + maskedbygeno[j][1] + maskedbygeno[j][2];
				double freq1 = (maskedbygeno[j][0] + maskedbygeno[j][1] * 0.5) / masked ;
				double maf = freq1;
				if (maf > 0.5) maf = 1.0 - maf;
				if (maskedbygeno[j][1] > 0)
				{
					double heterr = (mmbygeno[j][2] + mmbygeno[j][3]) * 0.5 / maskedbygeno[j][1];
					fprintf(of, " %5.4f %s %s %5.4f %5.4f %d %d %d %d %d %d %d %d %d", 
							heterr,
							(const char *) info->GetAlleleLabel(1), 
				                 	(const char *) info->GetAlleleLabel(2), 
							freq1, maf, 
							 maskedbygeno[j][0], mmbygeno[j][0], mmbygeno[j][1],
							 maskedbygeno[j][1], mmbygeno[j][2], mmbygeno[j][3],
							 maskedbygeno[j][2], mmbygeno[j][4], mmbygeno[j][5]);
				}
				else
				{
		                                        fprintf(of, " NA %s %s %5.4f %5.4f %d %d %d %d %d %d %d %d %d",
                                                        (const char *) info->GetAlleleLabel(1),
                                                        (const char *) info->GetAlleleLabel(2),
                                                        freq1, maf,
                                                         maskedbygeno[j][0], mmbygeno[j][0], mmbygeno[j][1],
                                                         maskedbygeno[j][1], mmbygeno[j][2], mmbygeno[j][3],
                                                         maskedbygeno[j][2], mmbygeno[j][4], mmbygeno[j][5]);
				}
			}

			if (accuracyByGeno)
			{
				int masked = maskedbygeno[j][0] + maskedbygeno[j][1] + maskedbygeno[j][2];
				double freq1 = (maskedbygeno[j][0] + maskedbygeno[j][1] * 0.5) / masked ;
				double hetaccuracy = 9.0, homo1accuracy = 9.0, homo2accuracy = 9.0; 
				if (maskedbygeno[j][1] > 0) {hetaccuracy = 1.0 - ((mmbygeno[j][2] + mmbygeno[j][3]) * 0.5 / maskedbygeno[j][1]);}
				// allelic
				if (maskedbygeno[j][0] > 0) {homo1accuracy = 1.0 - ((mmbygeno[j][0] + 2.0 * mmbygeno[j][1]) / maskedbygeno[j][0] / 2.0);}
				if (maskedbygeno[j][2] > 0) {homo2accuracy = 1.0 - ((mmbygeno[j][5] + 2.0 * mmbygeno[j][4]) / maskedbygeno[j][2] / 2.0);}
				// AL1 is the major allele
				if (freq1 >= 0.5)
				{
	                                fprintf(of, " %s %s %5.4f",
                                                        (const char *) info->GetAlleleLabel(1),
                                                        (const char *) info->GetAlleleLabel(2),
                                                        freq1);
					if (homo1accuracy == 9.0) {fprintf(of, " NA");}
					else {fprintf(of, " %5.4f", homo1accuracy);}

                                        if (hetaccuracy == 9.0) {fprintf(of, " NA");}
                                        else {fprintf(of, " %5.4f", hetaccuracy);}

                                        if (homo2accuracy == 9.0) {fprintf(of, " NA");}
                                        else {fprintf(of, " %5.4f", homo2accuracy);}
				}
				else
				{
					freq1 = 1.0 - freq1,
	                                fprintf(of, " %s %s %5.4f",
                                                        (const char *) info->GetAlleleLabel(2),
                                                        (const char *) info->GetAlleleLabel(1),
                                                        freq1);
					if (homo2accuracy == 9.0) {fprintf(of, " NA");}
					else {fprintf(of, " %5.4f", homo2accuracy);}

                                        if (hetaccuracy == 9.0) {fprintf(of, " NA");}
                                        else {fprintf(of, " %5.4f", hetaccuracy);}

                                        if (homo1accuracy == 9.0) {fprintf(of, " NA");}
                                        else {fprintf(of, " %5.4f", homo1accuracy);}

                                }

			}
			fprintf(of, "\n");
		}

		if(flippedmarkers.Find(truth.markerNames[j]) >= 0)
			fprintf(strandf, "%s\n", (const char *) truth.markerNames[j]);
	}
	
	printf("Summary: \n");
	printf("\tNumber of Compared Genotypes: %d\n", ttlmaskedgenos);
	
	if (ttlmaskedgenos > 0)
	{
		double overallGerr = 100.00 - (double) ttlmatchedgenos * 100.00 / (double) ttlmaskedgenos ;
		double overallAerr = 100.00 - (double) ttlmatchedalleles * 50.00 / (double) ttlmaskedgenos;
		printf("\tCompletely matched genotypes: %d\n", ttlmatchedgenos);
		printf("\tMatched alleles: %d\n", ttlmatchedalleles);
		printf("\tGenotype matching error: %.4f%% and allele matching error: %.4f%%\n", overallGerr, overallAerr);
	}

	printf("\n");

endt = clock();
int lapsetime = (int) ((double)(endt - startt) / CLOCKS_PER_SEC);      
printf("Analysis took %d seconds\n\n", lapsetime);

}







